//===========================================================================
/*
    This file is part of the CHAI 3D visualization and haptics libraries.
    Copyright (C) 2003-2011 by CHAI 3D. All rights reserved.

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License("GPL") version 2
    as published by the Free Software Foundation.

    For using the CHAI 3D libraries with software that can not be combined
    with the GNU GPL, and for taking advantage of the additional benefits
    of our support services, please contact CHAI 3D about acquiring a
    Professional Edition License.

    \author    <http://www.chai3d.org>
    \author    Francois Conti
    \version   2.2.0 $Rev: 566 $
 */
//===========================================================================

//---------------------------------------------------------------------------
#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
//---------------------------------------------------------------------------
#include "chai3d.h"
#include "CODE.h"
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// DECLARED CONSTANTS
//---------------------------------------------------------------------------

// initial size (width/height) in pixels of the display window
const int WINDOW_SIZE_W         = 512;
const int WINDOW_SIZE_H         = 512;

// mouse menu options (right button)
const int OPTION_FULLSCREEN     = 1;
const int OPTION_WINDOWDISPLAY  = 2;


//---------------------------------------------------------------------------
// DECLARED VARIABLES
//---------------------------------------------------------------------------

// a world that contains all objects of the virtual environment
cWorld* world;

// a camera that renders the world in a window display
cCamera* camera;

// a light source to illuminate the objects in the virtual scene
cLight *light;

// a little "chai3d" bitmap logo at the bottom of the screen
cBitmap* logo;

// width and height of the current window display
int displayW  = 0;
int displayH  = 0;

// stiffness of virtual spring between haptic device and tool (drill)
double linGain  = 0.20;
double angGain  = 0.03;
double linG     = 0.00;
double angG     = 0.00;
double linStiffness = 800;
double angStiffness = 30;

// a haptic device handler
cHapticDeviceHandler* handler;

// a virtual tool representing the haptic device in the scene
cGeneric3dofPointer* tool;

// radius of the tool proxy
double proxyRadius;

// maximum stiffness to be used with virtual objects in scene
double stiffnessMax;

// status of the main simulation haptics loop
bool simulationRunning = false;

// clock
cPrecisionClock startHapticsClock;

// ODE world
cODEWorld* ODEWorld;

// ODE object - teeth
cODEGenericBody* ODETeeth;

// ODE object - tool
cODEGenericBody* ODETool;

// Frame tip
cGenericObject* toolTip;

// root resource path
string resourceRoot;

// has exited haptics simulation thread
bool simulationFinished = false;


//---------------------------------------------------------------------------
// DECLARED MACROS
//---------------------------------------------------------------------------

// convert to resource path
#define RESOURCE_PATH(p)    (char*)((resourceRoot+string(p)).c_str())


//---------------------------------------------------------------------------
// DECLARED FUNCTIONS
//---------------------------------------------------------------------------

// callback when the window display is resized
void resizeWindow(int w, int h);

// callback when a keyboard key is pressed
void keySelect(unsigned char key, int x, int y);

// callback when the right mouse button is pressed to select a menu item
void menuSelect(int value);

// function called before exiting the application
void close(void);

// main graphics callback
void updateGraphics(void);

// main haptics loop
void updateHaptics(void);


//===========================================================================
/*
    DEMO:    ODE_mesh.cpp

    This example illustrates the use of the ODE framework for simulating
    haptic interaction with dynamic bodies. In this scene we demonstrate a 
    dental applicationwhere a drill is interacting with a set of teeth.

 */
//===========================================================================

int main(int argc, char* argv[])
{
    //-----------------------------------------------------------------------
    // INITIALIZATION
    //-----------------------------------------------------------------------

    printf ("\n");
    printf ("-----------------------------------\n");
    printf ("CHAI 3D\n");
    printf ("Demo: 43-ODE-dental\n");
    printf ("Copyright 2003-2011\n");
    printf ("-----------------------------------\n");
    printf ("\n\n");
    printf ("Instructions:\n\n");
    printf ("- Use haptic device and user switch to manipulate cubes \n");
    printf ("\n\n");
    printf ("Keyboard Options:\n\n");
	printf ("[1] - Enable gravity\n");
	printf ("[2] - Disable gravity\n");
	printf ("\n");
	printf ("[a] - decrease linear haptic gain\n");
	printf ("[s] - increase linear haptic gain\n");
	printf ("[d] - decrease angular haptic gain\n");
	printf ("[f] - increase angular haptic gain\n");
	printf ("\n");
	printf ("[g] - decrease linear stiffness\n");
	printf ("[h] - increase linear stiffness\n");
	printf ("[j] - decrease angular stiffness\n");
	printf ("[k] - increase angular stiffness\n");
	printf ("\n");
    printf ("[0] - Stereo    (ON/OFF)\n");
	printf ("[6] - Stereo Eye Separation (+)\n");
	printf ("[7] - Stereo Eye Separation (-)\n");
	printf ("[8] - Stereo Focal Length (+)\n");
	printf ("[9] - Stereo Focal Length (-)\n");
	printf ("\n");
    printf ("[x] - Exit application\n");
    printf ("\n\n");

    // parse first arg to try and locate resources
    resourceRoot = string(argv[0]).substr(0,string(argv[0]).find_last_of("/\\")+1);


    //-----------------------------------------------------------------------
    // 3D - SCENEGRAPH
    //-----------------------------------------------------------------------

    // create a new world.
    world = new cWorld();

    // set the background color of the environment
    // the color is defined by its (R,G,B) components.
    world->setBackgroundColor(0.0, 0.0, 0.0);

    // create a camera and insert it into the virtual world
    camera = new cCamera(world);
    world->addChild(camera);

    // position and oriente the camera
    camera->set( cVector3d (2.5, 0.0, 0.3),    // camera position (eye)
                 cVector3d (0.0, 0.0, 0.0),    // lookat position (target)
                 cVector3d (0.0, 0.0, 1.0));   // direction of the "up" vector

    // set the near and far clipping planes of the camera
    // anything in front/behind these clipping planes will not be rendered
    camera->setClippingPlanes(0.01, 10.0);

	// set stereo parameters
	camera->setUseStereo(true);
	camera->setStereoEyeSeparation(0.07);
	camera->setStereoFocalLength(2.1);

	// flip image horizontally for mirrored display
	camera->flipHorizontal(false);

    // create a light source and attach it to the camera
    light = new cLight(world);
    camera->addChild(light);                   // attach light to camera
    light->setEnabled(true);                   // enable light source
    light->setPos(cVector3d( 2.0, 0.5, 1.0));  // position the light source
    light->setDir(cVector3d(-2.0, 0.5, 1.0));  // define the direction of the light beam
    light->m_ambient.set(0.6, 0.6, 0.6);
    light->m_diffuse.set(0.8, 0.8, 0.8);
    light->m_specular.set(0.8, 0.8, 0.8);


    //-----------------------------------------------------------------------
    // 2D - WIDGETS
    //-----------------------------------------------------------------------

    // create a 2D bitmap logo
    logo = new cBitmap();

    // add logo to the front plane
    camera->m_front_2Dscene.addChild(logo);

    // load a "chai3d" bitmap image file
    bool fileload;
    fileload = logo->m_image.loadFromFile(RESOURCE_PATH("resources/images/chai3d.bmp"));
    if (!fileload)
    {
        #if defined(_MSVC)
        fileload = logo->m_image.loadFromFile("../../../bin/resources/images/chai3d.bmp");
        #endif
    }

    // position the logo at the bottom left of the screen (pixel coordinates)
    logo->setPos(10, 10, 0);

    // scale the logo along its horizontal and vertical axis
    logo->setZoomHV(0.25, 0.25);

    // here we replace all black pixels (0,0,0) of the logo bitmap
    // with transparent black pixels (0, 0, 0, 0). This allows us to make
    // the background of the logo look transparent.
    logo->m_image.replace(
        cColorb(0, 0, 0),      // original RGB color
        cColorb(0, 0, 0, 0)    // new RGBA color
    );

    // enable transparency
    logo->enableTransparency(true);


    //-----------------------------------------------------------------------
    // HAPTIC DEVICES / TOOLS
    //-----------------------------------------------------------------------

    // create a haptic device handler
    handler = new cHapticDeviceHandler();

    // get access to the first available haptic device
    cGenericHapticDevice* hapticDevice;
    handler->getDevice(hapticDevice, 0);

    // retrieve information about the current haptic device
    cHapticDeviceInfo info;
    if (hapticDevice)
    {
        info = hapticDevice->getSpecifications();
    }

    // create a 3D tool and add it to the world
    tool = new cGeneric3dofPointer(world);
    world->addChild(tool);

    // connect the haptic device to the tool
    tool->setHapticDevice(hapticDevice);

    // initialize tool by connecting to haptic device
    tool->start();

    // map the physical workspace of the haptic device to a larger virtual workspace.
    tool->setWorkspaceRadius(1.4);

    // define a radius for the tool (graphical display)
    tool->setRadius(0.015);

    // hide the device sphere. only show proxy.
    tool->m_deviceSphere->setShowEnabled(false);

    // set the physical readius of the proxy.
    proxyRadius = 0.0;
    tool->m_proxyPointForceModel->setProxyRadius(proxyRadius);
    tool->m_proxyPointForceModel->m_collisionSettings.m_checkBothSidesOfTriangles = false;

    // enable if objects in the scene are going to rotate of translate
    // or possibly collide against the tool. If the environment
    // is entirely static, you can set this parameter to "false"
    tool->m_proxyPointForceModel->m_useDynamicProxy = true;

    // ajust the color of the tool
    tool->m_materialProxy = tool->m_materialProxyButtonPressed;

    // read the scale factor between the physical workspace of the haptic
    // device and the virtual workspace defined for the tool
    double workspaceScaleFactor = tool->getWorkspaceScaleFactor();

    // define a maximum stiffness that can be handled by the current
    // haptic device. The value is scaled to take into account the
    // workspace scale factor
    stiffnessMax = info.m_maxForceStiffness / workspaceScaleFactor;


    //-----------------------------------------------------------------------
    // COMPOSE THE VIRTUAL SCENE
    //-----------------------------------------------------------------------

    // create an ODE world to simulate dynamic bodies
    ODEWorld = new cODEWorld(world);

    // add ODE world as a node inside world
    world->addChild(ODEWorld);

    // set some gravity
    ODEWorld->setGravity(cVector3d(0.0, 0.0, -0.5));

    // create a new ODE object that is automatically added to the ODE world
    ODETeeth = new cODEGenericBody(ODEWorld);

    // create a virtual mesh  that will be used for the geometry
    // representation of the dynamic body
    cMesh* imgTeeth = new cMesh(world);


    ////////////////////////////////////////////////////////////////////////////
    // LOAD TEETH MODEL
    ////////////////////////////////////////////////////////////////////////////

    if (true)
    {
        fileload = imgTeeth->loadFromFile(RESOURCE_PATH("resources/models/dental/teeth-bottom.3ds"));
        if (!fileload)
        {
            #if defined(_MSVC)
            fileload = imgTeeth->loadFromFile("../../../bin/resources/models/dental/teeth-bottom.3ds");
            #endif
        }
        imgTeeth->scale(0.004);
        imgTeeth->createAABBCollisionDetector(0.0, true, false);
        imgTeeth->setStiffness(300, true);

        // add mesh to ODE object
        ODETeeth->setImageModel(imgTeeth);

        // create a dynamic model of the ODE object.
        ODETeeth->createDynamicMesh(true);

        // position and oriente the ODE model
        cVector3d pos;
        cMatrix3d rot;
        pos = cVector3d(0.0, 0.0, -0.5);
        rot.identity();
        rot.rotate(cVector3d(0,0,1), cDegToRad(90));
        rot.rotate(cVector3d(0,1,0), cDegToRad(20));
        ODETeeth->setPosition(pos);
        ODETeeth->setRotation(rot);
    }
    else
    {
        fileload = imgTeeth->loadFromFile(RESOURCE_PATH("resources/models/dental/mandible.3ds"));
        if (!fileload)
        {
            #if defined(_MSVC)
            fileload = imgTeeth->loadFromFile("../../../bin/resources/models/dental/mandible.3ds");
            #endif
        }
        imgTeeth->scale(0.008);
        imgTeeth->createAABBCollisionDetector(0.0, true, false);
        imgTeeth->setStiffness(300, true);

        // add mesh to ODE object
        ODETeeth->setImageModel(imgTeeth);

        // create a dynamic model of the ODE object.
        ODETeeth->createDynamicMesh(true);

        // position and oriente the ODE model
        cVector3d pos;
        cMatrix3d rot;
        pos = cVector3d(0.0, 0.0, -0.1);
        rot.identity();
        rot.rotate(cVector3d(1,0,0), cDegToRad(180));
        rot.rotate(cVector3d(0,0,1), cDegToRad(90));
        rot.rotate(cVector3d(0,1,0), cDegToRad(15));
        ODETeeth->setPosition(pos);
        ODETeeth->setRotation(rot);
    }
 
	// create an ODE object that represents the tool
	ODETool = new cODEGenericBody(ODEWorld);
	cMesh* imgTool = new cMesh(world);

    ////////////////////////////////////////////////////////////////////////////
    // LOAD TOOL MODEL
    ////////////////////////////////////////////////////////////////////////////

    fileload = imgTool->loadFromFile(RESOURCE_PATH("resources/models/dental/drill.obj"));
    if (!fileload)
    {
        #if defined(_MSVC)
        fileload = imgTool->loadFromFile("../../../bin/resources/models/dental/drill.obj");
        #endif
    }
    imgTool->scale(0.01);

    // define some material properties for each cube
    cMaterial matTool;
    matTool.m_ambient.set(0.3, 0.3, 0.4);
    matTool.m_diffuse.set(0.6, 0.6, 0.7);
    matTool.m_specular.set(0.9, 0.9, 1.0);
    imgTool->setMaterial(matTool, true);
    imgTool->setHapticEnabled(false);
 
    // add mesh to ODE object
    ODETool->setImageModel(imgTool);
    ODETool->createDynamicMesh(false);

    // define some mass and damping properties for the tool 
    ODETool->setMass(0.01);
	dBodySetAngularDamping(ODETool->m_ode_body, 0.06);
	dBodySetLinearDamping(ODETool->m_ode_body, 0.06);

    // define an initial position of the ODE tool to be above the teeth model and not intersecting
    cVector3d pos(0.0, 0.0, 0.5);
    ODETool->setPosition(pos);

    // create a small object which represents the position of the tip of the drill
    toolTip = new cGenericObject();
    world->addChild(toolTip);
    toolTip->setShowFrame(true);
    toolTip->setFrameSize(0.2);
  
    //-----------------------------------------------------------------------
    // OPEN GL - WINDOW DISPLAY
    //-----------------------------------------------------------------------

    // initialize GLUT
    glutInit(&argc, argv);

    // retrieve the resolution of the computer display and estimate the position
    // of the GLUT window so that it is located at the center of the screen
    int screenW = glutGet(GLUT_SCREEN_WIDTH);
    int screenH = glutGet(GLUT_SCREEN_HEIGHT);
    int windowPosX = (screenW - WINDOW_SIZE_W) / 2;
    int windowPosY = (screenH - WINDOW_SIZE_H) / 2;

    // initialize the OpenGL GLUT window
    glutInitWindowPosition(windowPosX, windowPosY);
    glutInitWindowSize(WINDOW_SIZE_W, WINDOW_SIZE_H);
    glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE | GLUT_STEREO);
    glutCreateWindow(argv[0]);
    glutDisplayFunc(updateGraphics);
    glutKeyboardFunc(keySelect);
    glutReshapeFunc(resizeWindow);
    glutSetWindowTitle("CHAI 3D");

    // create a mouse menu (right button)
    glutCreateMenu(menuSelect);
    glutAddMenuEntry("full screen", OPTION_FULLSCREEN);
    glutAddMenuEntry("window display", OPTION_WINDOWDISPLAY);
    glutAttachMenu(GLUT_RIGHT_BUTTON);


    //-----------------------------------------------------------------------
    // START SIMULATION
    //-----------------------------------------------------------------------

    // simulation in now running
    simulationRunning = true;

    // create a thread which starts the main haptics rendering loop
    cThread* hapticsThread = new cThread();
    hapticsThread->set(updateHaptics, CHAI_THREAD_PRIORITY_HAPTICS);

    // start the main graphics rendering loop
    glutMainLoop();

    // close everything
    close();

    // exit
    return (0);
}

//---------------------------------------------------------------------------

void resizeWindow(int w, int h)
{
    // update the size of the viewport
    displayW = w;
    displayH = h;
    glViewport(0, 0, displayW, displayH);
}

//---------------------------------------------------------------------------

void keySelect(unsigned char key, int x, int y)
{
    // escape key
    if ((key == 27) || (key == 'x'))
    {
        // close everything
        close();

        // exit application
        exit(0);
    }

    // option 1:
    if (key == '1')
    {
        // enable gravity
        ODEWorld->setGravity(cVector3d(0.0, 0.0, -0.5));
    }

    // option 2:
    if (key == '2')
    {
        // disable gravity
        ODEWorld->setGravity(cVector3d(0.0, 0.0, 0.0));
    }

	//---------------------------------------------------------------
	// STEREO OPTIONS
	//---------------------------------------------------------------
    // enable/disable stereo:
    if (key == '0')
    {
        bool useStereo = camera->getUseStereo();
        camera->setUseStereo(!useStereo);
    }

    // increase eye separation
    if (key == '6')
    {
		double eyeSeparation = camera->getStereoEyeSeparation();
		eyeSeparation += 0.01;
        camera->setStereoEyeSeparation(eyeSeparation);
    }

    // reduce eye separation
    if (key == '7')
    {
		double eyeSeparation = camera->getStereoEyeSeparation();
		eyeSeparation -= 0.01;
        camera->setStereoEyeSeparation(eyeSeparation);
    }

    // increase focal length
    if (key == '8')
    {
		double focalLength = camera->getStereoFocalLength();
		focalLength += 0.1;
        camera->setStereoFocalLength(focalLength);
    }

    // reduce focal length
    if (key == '9')
    {
		double focalLength = camera->getStereoFocalLength();
		focalLength -= 0.1;
        camera->setStereoFocalLength(focalLength);
    }

	//---------------------------------------------------------------
	// HAPTIC OPTIONS
	//---------------------------------------------------------------
    // option a: decrease linear haptic gain
    if (key == 'a')
    {
        linGain = linGain - 0.05;
		if (linGain < 0)
			linGain = 0;
		printf("linear haptic gain:  %f\n", linGain);
    }

    // option s: increase linear haptic gain
    if (key == 's')
    {
        linGain = linGain + 0.05;
		printf("linear haptic gain:  %f\n", linGain);
    }

    // option d: decrease angular haptic gain
    if (key == 'd')
    {
        angGain = angGain - 0.005;
		if (angGain < 0)
			angGain = 0;
		printf("angular haptic gain:  %f\n", angGain);
    }

    // option f: increase angular haptic gain
    if (key == 'f')
    {
		angGain = angGain + 0.005;
		printf("angular haptic gain:  %f\n", angGain);
    }

    // option g: decrease linear stiffness
    if (key == 'g')
    {
        linStiffness = linStiffness - 50;
		if (linStiffness < 0)
			linStiffness = 0;
		printf("linear stiffness:  %f\n", linStiffness);
    }

    // option h: increase linear stiffness
    if (key == 'h')
    {
        linStiffness = linStiffness + 50;
		printf("linear stiffness:  %f\n", linStiffness);
    }

    // option j: decrease angular stiffness
    if (key == 'j')
    {
        angStiffness = angStiffness - 1;
		if (angStiffness < 0)
			angStiffness = 0;
		printf("angular stiffness:  %f\n", angStiffness);
    }

    // option k: increase angular stiffness
    if (key == 'k')
    {
        angStiffness = angStiffness + 1;
		printf("angular stiffness:  %f\n", angStiffness);
    }
}

//---------------------------------------------------------------------------

void menuSelect(int value)
{
    static int _wx, _wy, _ww, _wh;

    switch (value)
    {
        // enable full screen display
        case OPTION_FULLSCREEN:
            _wx = glutGet(GLUT_WINDOW_X);
            _wy = glutGet(GLUT_WINDOW_Y);
            _ww = glutGet(GLUT_WINDOW_WIDTH);
            _wh = glutGet(GLUT_WINDOW_HEIGHT);
            glutFullScreen();
            break;

        // reshape window to original size
        case OPTION_WINDOWDISPLAY:
            glutPositionWindow(_wx, _wy);
            glutReshapeWindow(_ww, _wh);
            break;
    }
}

//---------------------------------------------------------------------------

void close(void)
{
    // stop the simulation
    simulationRunning = false;

    // wait for graphics and haptics loops to terminate
    while (!simulationFinished) { cSleepMs(100); }

    // close haptic device
    tool->stop();
    }

//---------------------------------------------------------------------------

void updateGraphics(void)
{
    // render world
    camera->renderView(displayW, displayH);

    // Swap buffers
    glutSwapBuffers();

    // check for any OpenGL errors
    GLenum err;
    err = glGetError();
    if (err != GL_NO_ERROR) printf("Error:  %s\n", gluErrorString(err));

    // inform the GLUT window to call updateGraphics again (next frame)
    if (simulationRunning)
    {
        glutPostRedisplay();
    }
}

//---------------------------------------------------------------------------

void updateHaptics(void)
{
    // simulation clock
    cPrecisionClock simClock;
    simClock.start(true);

    // reset haptics activation clock
    startHapticsClock.reset();
	startHapticsClock.start();
	bool hapticsReady = false;
    double t = 0.0;

    // main haptic simulation loop
    while(simulationRunning)
    {
         // retrieve simulation time and compute next interval
        double time = simClock.getCurrentTimeSeconds();
        t = t + time;
        double nextSimInterval = 0.0005;//cClamp(time, 0.00001, 0.0002);
        
		// reset clock
        simClock.reset();
        simClock.start();

        // compute global reference frames for each object
        world->computeGlobalPositions(true);

        // update position and orientation of tool
        tool->updatePose();

        // compute interaction forces
        tool->computeInteractionForces();

        // update position and orientation of tool
		cVector3d posDevice;
		cMatrix3d rotDevice;
        posDevice = tool->getProxyGlobalPos();
        rotDevice = tool->getProxyGlobalRot();

		// read position of tool
		cVector3d posTool = ODETool->getPos();
		cMatrix3d rotTool = ODETool->getRot();

		// compute position and angular error between tool and haptic device
		cVector3d deltaPos = (posDevice - posTool);
		cMatrix3d deltaRot = cMul(cTrans(rotTool), rotDevice);
		double angle;
		cVector3d axis;
		deltaRot.toAngleAxis(angle, axis);
		
		// compute force and torque to apply to tool
		cVector3d force, torque;
		force = linStiffness * deltaPos;
		ODETool->addExternalForce(force);

		torque = cMul((angStiffness * angle), axis);
		rotTool.mul(torque);	
		ODETool->addExternalTorque(torque);
		
		// compute force and torque to apply to haptic device
		force = -linG * force;
		torque = -angG * torque;

        // add force contribution from ODE model
        tool->m_lastComputedGlobalForce.add(force);

        // send forces to haptic device.
        tool->applyForces();
		
        // gradually increase virtual linkage between haptic device and tool
        // at startup. 
		if (linG < linGain)
		{
			linG = linG + 0.1 * time * linGain;
		}
		else
		{
			linG = linGain;
		}

		if (angG < angGain)
		{
			angG = angG + 0.1 * time * angGain;
		}
		else
		{
			angG = angGain;
		}

        // update simulation
        ODEWorld->updateDynamics(nextSimInterval);

        // retrieve position of tooltip
        cMatrix3d rot = ODETool->getRot();
        cVector3d pos = cAdd(ODETool->getPos(), cMul(rot, cVector3d(0,0,0)));
       
        toolTip->setPos(pos);
        toolTip->setRot(rot);
    }

    // exit haptics thread
    simulationFinished = true;
}

//---------------------------------------------------------------------------
